import {getScreen} from "./engine/internal/screen.ts";
import {BG_OUTER} from "./colors.ts";
import {D, I} from "./engine/public.ts";
import {IGame, Point, Rect, Size} from "./engine/datatypes.ts";
import {HuntMode} from "./huntmode.ts";
import {getLayoutRect} from "./layout.ts";

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point;

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position = new Point(
      adjust(this.position.x, this.target.x),
      adjust(this.position.y, this.target.y),
    );
  }
}

type GameState = "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): Point {
  if (state === "Gameplay") {
    return new Point(0, 0);
  }
  if (state === "Thralls") {
    return new Point(0, 1);
  }

  throw `invalid state: ${state}`
}

export class Game implements IGame {
  camera: MenuCamera;
  state: GameState;
  huntMode: HuntMode;

  constructor() {
    this.camera = new MenuCamera({
      position: new Point(0, 0),
      target: new Point(0, 0),
    });
    this.state = "Gameplay";

    this.huntMode = HuntMode.generate({depth: 1});
  }

  update() {
    if (I.isKeyPressed("w")) {
      this.state = "Gameplay"
    }
    if (I.isKeyPressed("s")) {
      this.state = "Thralls"
    }

    this.camera.target = getScreenLocation(this.state);
    D.camera = new Point(
      D.size.w * this.camera.position.x,
      D.size.h * this.camera.position.y,
    )
    this.camera.update();

    // state-specific updates
    this.updateGameplay();
  }

  draw() {
    // draw screen background
    let oldCamera = D.camera;
    D.camera = new Point(0, 0);
    D.fillRect(new Point(0, 0), D.size, BG_OUTER);
    D.camera = oldCamera;

    this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let mouse = I.mousePosition?.offset(D.camera);
    if (mouse != null) {
      D.invertRect(mouse.offset(new Point(-1, -1)), new Size(3, 3))
    }
  }

  getGameStateRect(gameState: GameState): Rect {
    let baseRect: Rect;
    switch(gameState) {
      case "Gameplay":
      case "Thralls":
        baseRect = getLayoutRect(new Size(384, 384));
        break;
      default:
        throw `not sure what base rect to use: ${gameState}`
    }

    // now move that base to a reasonable screen location
    let screen = getScreen();
    let {w, h} = screen.size;
    let overallScreenLocation = getScreenLocation(gameState);

    let offsetX = overallScreenLocation.x * w;
    let offsetY = overallScreenLocation.y * h;

    return baseRect.offset(new Point(offsetX, offsetY));
  }

  #withHuntModeCamera(cb: () => void) {
    let region = this.getGameStateRect("Gameplay")

    D.withCamera(D.camera.offset(region.top.negate()), cb)
  }

  updateGameplay() {
    this.#withHuntModeCamera(() => {
      this.huntMode.update();
    });
  }

  drawGameplay() {
    this.#withHuntModeCamera(() => {
      this.huntMode.draw();
    });
  }
}


export let game = new Game();